"""
Types for the sizes returned by the ``flickr.photos.getSizes`` API.
"""

import typing


class PhotoSize(typing.TypedDict):
    """
    A variant of a photo you can download.

    This always has a width/height.
    """

    label: str
    width: int
    height: int
    media: typing.Literal["photo"]
    source: str


class VideoSize(typing.TypedDict):
    """
    A variant of a video you can download.

    This may not have a width/height, e.g. videos with the 'appletv' label.
    """

    label: str
    width: int | None
    height: int | None
    media: typing.Literal["video"]
    source: str


Size = PhotoSize | VideoSize



"""
Convert values from the Flickr API into nicely-typed values.
"""

import collections
from datetime import datetime, timezone
import re
import typing
from xml.etree import ElementTree as ET

from nitrate.xml import find_optional_text

from .models import (
    DateTaken,
    LocationContext,
    MachineTags,
    NumericLocation,
    NamedLocation,
    Rotation,
    SafetyLevel,
    TakenGranularity,
    User,
)


__all__ = [
    "create_user",
    "fix_realname",
    "parse_date_taken",
    "parse_machine_tags",
    "parse_numeric_location",
    "parse_named_location",
    "parse_safety_level",
    "parse_timestamp",
]


def parse_timestamp(ts: str, /) -> datetime:
    """
    Convert a Unix timestamp into a Python-native ``datetime``.

    Example:

        >>> parse_timestamp('1490376472')
        datetime(2017, 3, 24, 17, 27, 52, tzinfo=timezone.utc)

    The Flickr API frequently returns dates as Unix timestamps, for example:

    *   When you call ``flickr.photos.getInfo``, the ``<dates>`` element
        includes the upload and last update dates as a timestamp
    *   When you call ``flickr.people.getInfo`` for a user with Flickr Pro,
        the ``expires`` attribute is a numeric timestamp.

    In this case a Unix timestamp is "an unsigned integer specifying
    the number of seconds since Jan 1st 1970 GMT" [1].

    [1] https://www.flickr.com/services/api/misc.dates.html
    """
    return datetime.fromtimestamp(int(ts), tz=timezone.utc)


def _parse_date_taken_value(dt: str) -> datetime:
    """
    Convert a "date taken" string to a Python-native ``datetime``.

    Example:

        >>> _parse_date_taken_value('2017-02-17 00:00:00')
        datetime(2017, 2, 17, 0, 0)

    """
    # See https://www.flickr.com/services/api/misc.dates.html
    #
    #     The date taken should always be displayed in the timezone
    #     of the photo owner, which is to say, don't perform
    #     any conversion on it.
    #
    return datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")


def _parse_date_taken_granularity(g: str) -> TakenGranularity:
    """
    Converts a numeric granularity level in the Flickr API into a
    human-readable value.

    See https://www.flickr.com/services/api/misc.dates.html
    """
    lookup_table: dict[str, TakenGranularity] = {
        "0": "second",
        "4": "month",
        "6": "year",
        "8": "circa",
    }

    try:
        return lookup_table[g]
    except KeyError:
        raise ValueError(f"Unrecognised date granularity: {g}")


def parse_date_taken(
    *, value: str, granularity: str, unknown: bool
) -> DateTaken | None:
    """
    Parse a "date taken" value from the Flickr API.

    When you retrieve a photo with ``flickr.photos.getInfo``, the taken
    value is made of several parts:

        <dates … takengranularity="0" takenunknown="1" lastupdate="1705439827"/>

    This function converts these parts into nicely-typed objects.  This
    may be ``None`` if the date taken is unknown.
    """
    # We intentionally omit sending any 'date taken' information
    # to callers if it's unknown.
    #
    # There will be a value in the API response, but if the taken date
    # is unknown, it's defaulted to the date the photo was posted.
    # See https://www.flickr.com/services/api/misc.dates.html
    #
    # This value isn't helpful to callers, so we omit it.  This reduces
    # the risk of somebody skipping the ``unknown`` parameter and using
    # the value in the wrong place.
    if unknown:
        return None

    # This is a weird value I've seen returned on some videos; I'm
    # not sure what it means, but it's not something we can interpret
    # as a valid date, so we treat "date taken" as unknown even if
    # the API thinks it knows it.
    elif value.startswith("0000-"):
        return None

    else:
        return {
            "value": _parse_date_taken_value(value),
            "granularity": _parse_date_taken_granularity(granularity),
        }


def parse_safety_level(s: str) -> SafetyLevel:
    """
    Converts a numeric safety level ID in the Flickr API into
    a human-readable value.

    See https://www.flickrhelp.com/hc/en-us/articles/4404064206996-Content-filters
    """
    lookup_table: dict[str, SafetyLevel] = {
        "0": "safe",
        "1": "moderate",
        "2": "restricted",
    }

    try:
        return lookup_table[s]
    except KeyError:
        raise ValueError(f"Unrecognised safety level: {s}")


def parse_numeric_location(elem_with_location: ET.Element) -> NumericLocation | None:
    """
    Get location information about a photo.

    This takes an XML element with latitude/longitude/accuracy attributes;
    this can be a <location> element (on a single photo) or a <photo> element
    (on collection responses).
    """
    # The accuracy parameter in the Flickr API response tells us
    # the precision of the location information (15 November 2023):
    #
    #     Recorded accuracy level of the location information.
    #     World level is 1, Country is ~3, Region ~6, City ~11, Street ~16.
    #     Current range is 1-16.
    #
    # But some photos have an accuracy of 0!  It's unclear what this
    # means or how we should map this -- lower numbers mean less accurate,
    # so this location information might be completely useless.
    #
    # Discard it rather than risk propagating bad data.
    if elem_with_location.attrib["accuracy"] == "0":
        return None

    return {
        "latitude": float(elem_with_location.attrib["latitude"]),
        "longitude": float(elem_with_location.attrib["longitude"]),
        "accuracy": int(elem_with_location.attrib["accuracy"]),
    }


def parse_named_location(location_elem: ET.Element) -> NamedLocation:
    """
    Get named location information about a photo.
    """
    # The context tells us whether the photo is indoors or outdoors.
    # See https://www.flickr.com/services/api/flickr.photos.geo.setLocation.html
    context_lookup: dict[str, LocationContext | None] = {
        "0": None,
        "1": "indoors",
        "2": "outdoors",
    }

    try:
        context = context_lookup[location_elem.attrib["context"]]
    except KeyError:  # pragma: no cover
        raise ValueError(
            f"Unrecognised location conext: {location_elem.attrib['context']}"
        )

    # On responses form the `flickr.photos.getInfo` endpoint, the
    # <location> element will be of the form:
    #
    #     <location latitude="9.135158" longitude="40.083811" accuracy="16" context="0">
    #       <locality>Galoch</locality>
    #       <neighbourhood/>
    #       <region>Āfar</region>
    #       <country>Ethiopia</country>
    #     </location>
    #
    return {
        "context": context,
        "locality": find_optional_text(location_elem, path="locality"),
        "county": find_optional_text(location_elem, path="county"),
        "region": find_optional_text(location_elem, path="region"),
        "country": find_optional_text(location_elem, path="country"),
        # The mismatched spelling here is intentional --
        # the Flickr API uses the British English spelling, but
        # we use US English throughout Data Lifeboat so we fix it.
        "neighborhood": find_optional_text(location_elem, path="neighbourhood"),
    }


def create_user(
    user_id: str,
    username: str,
    realname: str | None,
    path_alias: str | None,
    is_deleted: bool = False,
) -> User:
    """
    Given some core attributes, construct a ``User`` object.

    This function is only intended for internal user.
    """
    realname = fix_realname(user_id, username=username, realname=realname)

    # The Flickr API is a bit inconsistent about how some undefined attributes
    # are returned, e.g. ``realname`` can sometimes be null, sometimes an
    # empty string.
    #
    # In our type system, we want all of these empty values to map to ``None``.
    user: User = {
        "id": user_id,
        "username": username,
        "realname": realname,
        "path_alias": path_alias or None,
        "photos_url": f"https://www.flickr.com/photos/{path_alias or user_id}/",
        "profile_url": f"https://www.flickr.com/people/{path_alias or user_id}/",
    }

    if is_deleted:
        return {**user, "is_deleted": True}
    else:
        return user


def fix_realname(user_id: str, username: str, realname: str | None) -> str | None:
    """
    Override the ``realname`` returned by the Flickr API.

    In general we should avoid adding too many fixes here because it would
    quickly get unwieldy, but it's a useful place to consolidate these
    fixes for members we work with a lot.
    """
    realname = realname or None

    # The museum removed the 'S' so it would look good with the big 'S'
    # in their buddy icon, but that doesn't work outside Flickr.com.
    #
    # This name needed fixing on 23 July 2024; if they ever change
    # the name on the actual account, we can remove this fudge.
    if user_id == "62173425@N02" and realname == "tockholm Transport Museum":
        return "Stockholm Transport Museum"

    # This is a frequent commenter on Flickr Commons photos.  There's a
    # realname present if you visit their profile on Flickr.com,
    # but it isn't returned in the API.
    #
    # This name needed fixing on 7 August 2024 and I've reported it
    # as an API bug; if it ever gets fixed, we can remove this branch.
    if user_id == "32162360@N00" and username == "ɹǝqɯoɔɥɔɐǝq" and realname is None:
        return "beachcomber australia"

    return realname


# The best documentation I can find for Flickr's implementation of
# machine tags is a group post from a Flickr staff member in 2007:
# https://www.flickr.com/groups/51035612836@N01/discuss/72157594497877875/
#
# A machine tag is made of three parts:
#
#     {namespace}:{predicate}={value}
#
# The ``namespace`` and ``predicate`` can only contain ASCII characters.
MACHINE_TAG_RE = re.compile(
    r"""
    ^
    (?P<namespace>[a-zA-Z0-9_]+)
    :
    (?P<predicate>[a-zA-Z0-9_]+)
    =
    (?P<value>.+)
    $
    """,
    re.VERBOSE,
)


def parse_machine_tags(tags: list[str]) -> MachineTags:
    """
    Given a list of raw tags on Flickr, parse the machine tags
    as key-value pairs.

    The keys are namespace/predicate; the values are values.

    Example:

        >>> tags = [
        ...     "square",
        ...     "shape:sides=4",
        ...     "shape:color=red",
        ...     "shape:color=blue"
        ... ]
        ...
        >>> get_machine_tags(tags)
        {"shape:sides": ["4"], "shape:color": ["red", "blue"]}

    This function is a "best effort" parsing of machine tags -- it may
    not match Flickr perfectly, but is meant to make it easier for
    callers to work with machine tags.
    """
    result: MachineTags = collections.defaultdict(list)

    for t in tags:
        if m := MACHINE_TAG_RE.match(t):
            namespace = m.group("namespace")
            predicate = m.group("predicate")
            value = m.group("value")

            result[f"{namespace}:{predicate}"].append(value)

    return dict(result)


def parse_rotation(rs: str) -> Rotation:
    """
    Given a raw rotation value from Flickr, turn it into a rotation.
    """
    if rs not in {"0", "90", "180", "270"}:  # pragma: no cover
        raise ValueError(f"Unrecognised rotation value: {rs!r}")

    return typing.cast(Rotation, int(rs))



"""
Methods for getting information about licenses from the Flickr API.
"""

import functools
import itertools
import re
import typing

from .base import FlickrApi
from ..exceptions import LicenseNotFound, ResourceNotFound
from ..models import License, LicenseChange, LicenseId
from ..models.licenses import LicenseChangeEntry
from ..parsers import parse_timestamp


class LicenseMethods(FlickrApi):
    """
    License-related methods for the Flickr API.
    """

    # Note: this list of licenses almost never changes, so we call this once
    # and cache the result for efficiency.
    @functools.cache
    def get_licenses(self) -> dict[str, License]:
        """
        Returns a list of licenses, organised by numeric ID.

        In particular, IDs can be looked up using the numeric ID
        returned by many Flickr API methods.

        See https://www.flickr.com/services/api/flickr.photos.licenses.getInfo.htm
        """
        license_resp = self.call(method="flickr.photos.licenses.getInfo")

        result: dict[str, License] = {}

        # Add a short ID which can be used to more easily refer to this
        # license throughout the codebase.
        license_ids: dict[str, LicenseId] = {
            "All Rights Reserved": "all-rights-reserved",
            "Attribution-NonCommercial-ShareAlike License": "cc-by-nc-sa-2.0",
            "Attribution-NonCommercial License": "cc-by-nc-2.0",
            "Attribution-NonCommercial-NoDerivs License": "cc-by-nc-nd-2.0",
            "Attribution License": "cc-by-2.0",
            "Attribution-ShareAlike License": "cc-by-sa-2.0",
            "Attribution-NoDerivs License": "cc-by-nd-2.0",
            "No known copyright restrictions": "nkcr",
            "United States Government Work": "usgov",
            "Public Domain Dedication (CC0)": "cc0-1.0",
            "Public Domain Mark": "pdm",
        }

        license_labels = {
            "Attribution-NonCommercial-ShareAlike License": "CC BY-NC-SA 2.0",
            "Attribution-NonCommercial License": "CC BY-NC 2.0",
            "Attribution-NonCommercial-NoDerivs License": "CC BY-NC-ND 2.0",
            "Attribution License": "CC BY 2.0",
            "Attribution-ShareAlike License": "CC BY-SA 2.0",
            "Attribution-NoDerivs License": "CC BY-ND 2.0",
            "Public Domain Dedication (CC0)": "CC0 1.0",
        }

        for lic in license_resp.findall(".//license"):
            result[lic.attrib["id"]] = {
                "id": license_ids[lic.attrib["name"]],
                "label": license_labels.get(lic.attrib["name"], lic.attrib["name"]),
                "url": lic.attrib["url"],
            }

        return result

    @functools.cache
    def lookup_license_by_id(self, *, id: str) -> License:
        """
        Return the license for a license ID.

        The ID can be one of:

        *   The numeric license ID returned from the Flickr API
            (e.g. "0" ~> "All Rights Reserved")
        *   The human-readable license ID returned from this library
            (e.g. "cc-by-2.0" ~> "CC BY 2.0")

        """
        licenses = self.get_licenses()

        # If this is a numeric ID, then it must have come from the
        # Flickr API.  Look it up directly in the dict.
        if re.match(r"^[0-9]+$", id):
            try:
                return licenses[id]
            except KeyError:
                raise LicenseNotFound(license_id=id)

        # Otherwise, this is a human-readable license ID from our
        # library, so look for a matching license.
        try:
            matching_license = next(lic for lic in licenses.values() if lic["id"] == id)
            return matching_license
        except StopIteration:
            raise LicenseNotFound(license_id=id)

    def get_license_history(self, photo_id: str) -> list[LicenseChange]:
        """
        Return the license history of a photo.

        This always returns license events in sorted order.
        """
        licenses_by_url = {lic["url"]: lic for lic in self.get_licenses().values()}

        # First call the getLicenseHistory API.
        # See https://www.flickr.com/services/api/flickr.photos.licenses.getLicenseHistory.html
        history_resp = self.call(
            method="flickr.photos.licenses.getLicenseHistory",
            params={"photo_id": photo_id},
            exceptions={"1": ResourceNotFound()},
        )

        # Look for <license_history> elements in the response.
        history_elems = history_resp.findall("./license_history")

        # If there's a single <license_history> element and the `new_license`
        # is empty, it means this is the original license.
        #
        #     <rsp stat="ok">
        #       <license_history
        #         date_change="1733215279"
        #         old_license="All Rights Reserved" old_license_url="https://www.flickrhelp.com/hc/en-us/articles/10710266545556-Using-Flickr-images-shared-by-other-members"
        #         new_license="" new_license_url=""
        #       />
        #     </rsp>
        #
        if len(history_elems) == 1 and history_elems[0].attrib["new_license"] == "":
            date_posted = parse_timestamp(history_elems[0].attrib["date_change"])
            license_url = history_elems[0].attrib["old_license_url"]

            return [
                {
                    "date_posted": date_posted,
                    "license": licenses_by_url[license_url],
                }
            ]

        # Restructure the <license_history> elements -- at this point,
        # we know that they all have both an `old_license` and a `new_license`.
        #
        # While we're here, let's make sure the events are in date order.
        # The Flickr API usually returns them in this order, but it's
        # not guaranteed -- let's make sure that's true.
        license_events: list[LicenseChangeEntry.ChangedLicense] = sorted(
            [
                {
                    "date_changed": parse_timestamp(elem.attrib["date_change"]),
                    "old_license": licenses_by_url[elem.attrib["old_license_url"]],
                    "new_license": licenses_by_url[elem.attrib["new_license_url"]],
                }
                for elem in history_elems
            ],
            key=lambda ev: ev["date_changed"],
        )

        # Do a quick consistency check that this history makes sense
        # -- when a license changes, the `old_license` is the same
        # as the previous `new_license`.
        for ev1, ev2 in itertools.pairwise(license_events):
            assert ev1["new_license"] == ev2["old_license"]

        return typing.cast(list[LicenseChange], license_events)



"""
Models for fields you get on photo objects.
"""

import typing


# When somebody uploads a photo to Flickr, they can choose to rotate it.
#
# As of April 2025, there are only four rotation options.
Rotation = typing.Literal[0, 90, 180, 270]


class NumericLocation(typing.TypedDict):
    """
    Coordinates for a location.
    """

    latitude: float
    longitude: float
    accuracy: int


LocationContext = typing.Literal["indoors", "outdoors"]


class NamedLocation(typing.TypedDict):
    """
    Human-readable names for a location.
    """

    context: LocationContext | None
    neighborhood: str | None
    locality: str | None
    county: str | None
    region: str | None
    country: str | None


class Location(NumericLocation, NamedLocation):
    """
    Both numeric and named information about a location.
    """


class Editability(typing.TypedDict):
    """
    Describes if/how somebody can interact with a photo.
    """

    can_comment: bool
    can_add_meta: bool


class Usage(typing.TypedDict):
    """
    Describes the permissions that can be set on a photo.
    """

    can_download: bool
    can_blog: bool
    can_print: bool
    can_share: bool


class ExifTag(typing.TypedDict):
    """
    An EXIF/TIFF/GPS tag for a photo, as returned by the
    `flickr.photos.getExif` API.

    See https://www.flickr.com/services/api/flickr.photos.getExif.html
    """

    tagspace: str
    tagspaceid: str
    tag: str
    label: str
    raw_value: str
    clean_value: typing.NotRequired[str]



"""
Base classes for the Flickr API client.
"""

import abc
from collections.abc import Mapping
import typing
from xml.etree import ElementTree as ET

import httpx
from nitrate.xml import find_required_elem
from tenacity import (
    retry,
    retry_if_exception,
    RetryError,
    stop_after_attempt,
    wait_random_exponential,
)

from ..exceptions import (
    InvalidApiKey,
    InvalidXmlException,
    UnrecognisedFlickrApiException,
)


HttpMethod = typing.Literal["GET", "POST"]


class FlickrApi(abc.ABC):
    """
    This is a basic model for Flickr API implementations: they have to provide
    a ``call()`` method that takes a Flickr API method and parameters, and returns
    the parsed XML.

    We deliberately split out the interface and implementation here -- currently
    we use httpx and tenacity, but this abstraction would allow us to swap out
    the underlying HTTP framework easily if we wanted to.
    """

    @abc.abstractmethod
    def call(
        self,
        *,
        http_method: HttpMethod = "GET",
        method: str,
        params: Mapping[str, str | int] | None = None,
        exceptions: dict[str, Exception] | None = None,
    ) -> ET.Element:
        """
        Call the Flickr API and return the XML of the result.

        :param method: The name of the Flickr API method, for example
            ``flickr.photos.getInfo``

        :param params: Any arguments to pass to the Flickr API method,
            for example ``{"photo_id": "1234"}``

        :param exceptions: A map from Flickr API error code to exceptions that should
            be thrown.
        """
        raise NotImplementedError


def is_retryable(exc: BaseException) -> bool:
    """
    Returns True if this is an exception we can safely retry (i.e. flaky
    or transient errors that might return a different result),or
    False otherwise.
    """
    if isinstance(exc, httpx.HTTPStatusError) and exc.response.status_code >= 500:
        return True

    if isinstance(exc, httpx.ReadTimeout):
        return True

    if isinstance(exc, InvalidXmlException):
        return True

    # These are a particular set of slightly flaky errors that are hard
    # to reproduce -- relying on matching the text of the error is a bit
    # fragile, but that's all we get from httpx.
    if isinstance(exc, (httpx.ReadError, httpx.ConnectError)) and exc.args == (
        "[Errno 54] Connection reset by peer",
    ):
        return True

    if isinstance(exc, httpx.RemoteProtocolError) and exc.args == (
        "Server disconnected without sending a response.",
    ):
        return True

    # Sometimes we get an error from the Flickr API like:
    #
    #     <err
    #       code="201"
    #       msg="Sorry, the Flickr API service is not currently available."
    #     />
    #
    # but this indicates a flaky connection rather than a genuine failure.
    #
    # We've seen similar with code "0", so we match on the error message
    # rather than the code.
    if (
        isinstance(exc, UnrecognisedFlickrApiException)
        and isinstance(exc.args[0], dict)
        and exc.args[0].get("msg")
        == "Sorry, the Flickr API service is not currently available."
    ):
        return True

    return False


class HttpxImplementation(FlickrApi):
    """
    An implementation of the Flickr API that uses ``httpx`` to make HTTP calls,
    and ``tenacity`` for retrying failed API calls.
    """

    def __init__(self, *, client: httpx.Client) -> None:
        """
        Create an API from an ``httpx`` client.

        This is useful if you want to customise the behaviour of the
        underlying client.
        """
        client.base_url = httpx.URL("https://api.flickr.com/services/rest/")
        self.client = client

    @classmethod
    def with_api_key(cls, *, api_key: str, user_agent: str) -> typing.Self:
        """
        Create a client from a Flickr API key.

        This also requires a User-Agent, which is a recommended good
        practice with the Flickr API (tho not required).
        """
        if not api_key:
            raise ValueError(
                "Cannot create a client with an empty string as the API key"
            )

        client = httpx.Client(
            params={"api_key": api_key},
            headers={"User-Agent": user_agent},
        )

        return cls(client=client)

    def call(
        self,
        *,
        http_method: HttpMethod = "GET",
        method: str,
        params: Mapping[str, str | int] | None = None,
        exceptions: dict[str, Exception] | None = None,
    ) -> ET.Element:
        """
        Call the Flickr API and return the XML of the result.

        :param method: The name of the Flickr API method, for example
            ``flickr.photos.getInfo``

        :param params: Any arguments to pass to the Flickr API method,
            for example ``{"photo_id": "1234"}``

        :param exceptions: A map from Flickr API error code to exceptions that should
            be thrown.
        """
        try:
            return self._call_api(
                http_method=http_method,
                method=method,
                params=params,
                exceptions=exceptions or {},
            )
        except RetryError as retry_err:
            retry_err.reraise()

    @retry(
        retry=retry_if_exception(is_retryable),
        stop=stop_after_attempt(5),
        wait=wait_random_exponential(),
    )
    def _call_api(
        self,
        *,
        http_method: HttpMethod,
        method: str,
        params: Mapping[str, str | int] | None,
        exceptions: dict[str, Exception],
    ) -> ET.Element:
        """
        Call the Flickr API and return the XML of the result.

        This function may be retried if the Flickr API returns an error
        that we think is retryable, e.g. if it returns
        a 500 Internal Server Erorr.
        """
        if params is not None:
            req_params = {"method": method, **params}
        else:
            req_params = {"method": method}

        resp = self.client.request(
            method=http_method, url="", params=req_params, timeout=15
        )
        resp.raise_for_status()

        # Note: the xml.etree.ElementTree is not secure against maliciously
        # constructed data (see warning in the Python docs [1]), but that's
        # fine here -- we're only using it for responses from the Flickr API,
        # which we trust.
        #
        # However, on occasion I have seen it return error messages in
        # JSON rather than XML, which causes this method to fail -- make
        # sure we log the offending text, and allow it to be retried as
        # a temporary failure.
        #
        # [1]: https://docs.python.org/3/library/xml.etree.elementtree.html
        try:
            xml = ET.fromstring(resp.text)
        except ET.ParseError as err:
            raise InvalidXmlException(
                f"Unable to parse response as XML ({resp.text!r}), got error {err}"
            )

        # If the Flickr API call fails, it will return a block of XML like:
        #
        #       <rsp stat="fail">
        #       	<err
        #               code="1"
        #               msg="Photo &quot;1211111111111111&quot; not found (invalid ID)"
        #           />
        #       </rsp>
        #
        # Different API endpoints have different codes, and so we just throw
        # and let calling functions decide how to handle it.
        if xml.attrib["stat"] == "fail":
            errors = find_required_elem(xml, path=".//err").attrib

            if errors["code"] == "100":
                raise InvalidApiKey(message=errors["msg"])

            try:
                raise exceptions[errors["code"]]
            except KeyError:
                # Note: the `from None` means we don't include the KeyError in
                # the traceback -- this is to avoid exposing internal details
                # of the library to external callers.
                raise UnrecognisedFlickrApiException(errors) from None

        return xml



from datetime import datetime
import typing

from .contexts import AlbumContext, GalleryContext, GroupContext, PhotoContext
from .licenses import License, LicenseId, LicenseChange
from .machine_tags import MachineTags
from .photo import (
    Editability,
    ExifTag,
    Location,
    LocationContext,
    NamedLocation,
    NumericLocation,
    Rotation,
    Usage,
)
from .sizes import Size
from .users import User, UserInfo


__all__ = [
    "AlbumContext",
    "BoundingBox",
    "Comment",
    "CommonsInstitution",
    "DateTaken",
    "Editability",
    "ExifTag",
    "GalleryContext",
    "GroupContext",
    "License",
    "LicenseId",
    "LicenseChange",
    "Location",
    "LocationContext",
    "MachineTags",
    "MediaType",
    "NamedLocation",
    "Note",
    "NumericLocation",
    "Person",
    "PhotoContext",
    "Rotation",
    "SafetyLevel",
    "SinglePhotoInfo",
    "SinglePhoto",
    "Size",
    "Tag",
    "TakenGranularity",
    "Usage",
    "User",
    "UserInfo",
    "Visibility",
]


# Represents the accuracy to which we know a date taken to be true.
#
# See https://www.flickr.com/services/api/misc.dates.html
TakenGranularity = typing.Literal["second", "month", "year", "circa"]


class DateTaken(typing.TypedDict):
    value: datetime
    granularity: TakenGranularity


class Comment(typing.TypedDict):
    """
    A comment as received from the Flickr API.
    """

    id: str
    photo_id: str
    author: User
    text: str
    permalink: str
    date: datetime


class Tag(typing.TypedDict):
    raw_value: str
    normalized_value: str

    author_id: str
    author_name: str

    is_machine_tag: bool


class Visibility(typing.TypedDict):
    is_public: bool
    is_friend: bool
    is_family: bool


# Represents the safety level of a photo on Flickr.
#
# https://www.flickrhelp.com/hc/en-us/articles/4404064206996-Content-filters#h_01HBRRKK6F4ZAW6FTWV8BPA2G7
SafetyLevel = typing.Literal["safe", "moderate", "restricted"]


MediaType = typing.Literal["photo", "video"]


class BoundingBox(typing.TypedDict):
    """
    A "bounding" box that highlights a specific region of a photo.
    """

    x: int
    y: int
    width: int
    height: int


class Person(typing.TypedDict):
    """
    A person tagged in a photo.
    """

    user: User
    bounding_box: BoundingBox | None


class Note(typing.TypedDict):
    """
    A note left on a photo.
    """

    id: str
    author: User
    bounding_box: BoundingBox
    text: str


class SinglePhotoInfo(typing.TypedDict):
    """
    Represents a response from the flickr.photos.getInfo API.
    """

    id: str
    media: MediaType

    secret: str
    server: str
    farm: str
    original_format: str | None

    rotation: Rotation

    owner: User

    safety_level: SafetyLevel

    license: License

    title: str | None
    description: str | None
    tags: list[str]
    machine_tags: MachineTags
    raw_tags: list[Tag]
    notes: list[Note]

    date_posted: datetime
    date_taken: DateTaken | None
    location: Location | None

    count_comments: int
    count_views: int
    has_people: bool

    visibility: Visibility
    editability: Editability
    public_editability: Editability
    usage: Usage

    url: str


class SinglePhoto(SinglePhotoInfo):
    sizes: list[Size]


class CommonsInstitution(typing.TypedDict):
    """
    Represents an institution in the Flickr Commons programme.
    """

    user_id: str
    date_launch: datetime
    name: str
    site_url: str | None
    license_url: str
